@@ -11,17 +11,15 @@ from TaskGen import extension, taskgen, after, before
 EXT_ASM = ['.s', '.S', '.asm', '.ASM', '.spp', '.SPP']
 
 as_str = '${AS} ${ASFLAGS} ${_ASINCFLAGS} ${SRC} -o ${TGT}'
-Task.simple_task_type('asm', as_str, 'PINK', ext_out='.o')
+Task.simple_task_type('asm', as_str, 'PINK', ext_out='.o', shell=False)
 
 @extension(EXT_ASM)
 def asm_hook(self, node):
 	# create the compilation task: cpp or cc
-	task = self.create_task('asm')
 	try: obj_ext = self.obj_ext
 	except AttributeError: obj_ext = '_%d.o' % self.idx
 
-	task.inputs = [node]
-	task.outputs = [node.change_ext(obj_ext)]
+	task = self.create_task('asm', node, node.change_ext(obj_ext))
 	self.compiled_tasks.append(task)
 	self.meths.append('asm_incflags')
 
@@ -35,12 +33,6 @@ def asm_incflags(self):
 	else: self.env['_ASINCFLAGS'] = self.env['_CCINCFLAGS']
 
 def detect(conf):
-	comp = os.environ.get('AS', '')
-	if not comp: comp = conf.find_program('as', var='AS')
-	if not comp: comp = conf.find_program('gas', var='AS')
-	if not comp: comp = conf.env['CC']
-	if not comp: return
-
-	v = conf.env
-	v['ASFLAGS']  = ''
+	conf.find_program(['gas', 'as'], var='AS')
+	if not conf.env.AS: conf.env.AS = conf.env.CC
 
